import numpy as np

def haversine(loc1, loc2):
    '''Compute the harversine distance between two points.
    Parameters
    ----------
        loc1: ndarray
            shape N*2, N pairs of (lon, lat).
        loc2: ndarray
            shape M*2, M pairs of (lon, lat).
    Return
    ------
        ndarray
            shape N*M, [i,j] denote the distance between loc1[i] and loc2[j]. in unit km. 
    '''
    loc1 = np.array(loc1)/180.0*np.pi
    loc2 = np.array(loc2)/180.0*np.pi
    R = 6378.137
    result = []
    for i in range(loc1.shape[0]):
        h = np.sin((loc1[i,1]-loc2[:,1])/2)**2 + np.cos(loc1[i,1])*np.cos(loc2[:,1])*np.sin((loc1[i,0]-loc2[:,0])/2)**2
        d = 2*R*np.arcsin(np.sqrt(h))
        result.append(d)
    return np.array(result)

def IDW(sloc, svalue, tloc, scale=0.05, kernel='exp'):
    '''The Inverse Distance Weighted interplotation method.
    Parameters
    ----------
        sloc: ndarray
            shape N*2, N pairs of scource location (lon, lat)
        svalue: ndarray
            shape N*K, corresponding scource value (or vector length K).
        tloc: ndarray
            shape M*2, M pairs of target location (lon, lat)
        scale: float, >0, default 0.1
            determine the influence range of a scource, the larger the larger.
        kernel: str, default 'exp'
            the inverse kernel function, optional: 'exp', '1/x'
    Return
    ------
        ndarray
            shape M*K, corresponding target value.
    '''
    kernel = {'exp': np.exp, '1/x': lambda x: 1/(x+1e-5)}[kernel]
    
    distance = haversine(sloc, tloc)
    distance_std = np.std(distance)
    if distance_std==0:
        # This happens when there are only one scource, we return a constant result in this case
        result = np.ones((tloc.shape[0],svalue.shape[1]))*svalue[0]
        return result
    weight = []
    for i in range(sloc.shape[0]):
        _w = kernel(-distance[i,:]/(distance_std*scale))
        weight.append(_w)
    weight = np.array(weight) + 1e-99
    weight = weight/np.sum(weight,0)
    result = np.matmul(svalue.T, weight).T
    return result


import pandas as pd
import numpy as np

def linear_normalize(datapack, columns='data0', method='99pt'):
    '''Normalize data linealy column-wise.
    Parameter
    ---------
        datapack: DataPack
        columns: str or list, default: 'data0'
            The column names to be normalized. if given 'ALL', all field start with 'data' will be processed.
        method: str, default: '99pt'
            The method used to calculate zoom boundary.
            Supported: 'max-min'(max-min), '99pt'(99 percentail value), '3std'(3 times of std range)
    Return
    ------
        DataPack
    '''
    if columns=='ALL':
        columns = [str(x) for x in datapack.data.columns if str(x).startswith('data')]
    elif type(columns) is str:
        columns = [columns]
    
    datapack.normalize_factor = {}
    for c in columns:
        series = datapack.data[c].values
        if method=='max-min':
            series_max = np.max(series)
            series_min = np.min(series)
        if method=='99pt':
            series_max = np.percentile(series, 99)
            series_min = np.percentile(series, 1)
        if method=='999pt':
            series_max = np.percentile(series, 99.9)
            series_min = np.percentile(series, 0.1)
        if method=='3std':
            series_max = np.mean(series) + 3 * np.std(series)
            series_min = series_max - 6 * np.std(series)
        if series_max==series_min:
            series_max = series_min + 1
        series = (series - series_min) / (series_max - series_min)
        datapack.data[c] = series
        datapack.normalize_factor.update({c: (series_min, series_max)})

    return datapack




from apdt.general import DataPack
import numpy as np
import pandas as pd
import yaml
import time
import json
import os
import requests

from apdt.config import _config

def get_weather_history(t,lon,lat,feature='temperature',secret_key=None):
    '''Get weather data in the past. This function will have 1 darksky API use. 
    Parameters
    ----------
        t: str, a date string like '2017-01-01'. It will be considered as UTC+8 (Beijing) time.
        lon: float, the target longtitude.
        lat: float, the target latitude.
        feature: str or list of str, the desired feature list.
        secret_key: str, a darksky secret key. if None is given, darksky_secret_key_0 in config file will be tried.
    Return
    ------
        ndarray, of shape 24*D, where D is the number of feature.
    '''

    # Parameter Check
    if not os.path.exists('data/.temp/'):
        os.mkdir('data/.temp/')
    if pd.to_datetime(t) > pd.to_datetime((time.time()+8*3600)*1e9):
        raise Exception('Input date should in history: '+t)
    if type(feature) is str:
        feature = [feature]
    alia_dict = {}
    alia_dict.update({x: 'temperature' for x in ['temperature','temp','tmp','wendu']})
    alia_dict.update({x: 'humidity' for x in ['humidity','hmd','hum','shidu']})
    alia_dict.update({x: 'windSpeed' for x in ['windSpeed','speed','spd','fengsu']})
    alia_dict.update({x: 'windBearing' for x in ['windBearing','direction','angel','fengxiang']})
    alia_dict.update({x: 'visibility' for x in ['visibility','kejiandu','keshidu']})
    alia_dict.update({x: 'pressure' for x in ['pressure','press','yali','qiya']}) 
    for i,x in enumerate(feature):
        if x not in alia_dict.keys():
            raise Exception(x+' is not supported.')
        feature[i] = alia_dict[x]
    if secret_key is None:
        if 'darksky_secret_key_0' in _config:
            secret_key = _config['darksky_secret_key_0']
        else:
            raise Exception('Secret key is not provide.')

    time_stamp = time.mktime((time.strptime(t+' 00:00:01','%Y-%m-%d %H:%M:%S')))
    req_loc = '{0:.2f},{1:.2f},'.format(lat,lon) + str(round(time_stamp))
    req = 'https://api.darksky.net/forecast/'+secret_key+'/'+req_loc
    file = 'data/.temp/'+req_loc+'.json'
    if os.path.exists(file):
        with open(file, 'r') as result_file:
            data = json.load(result_file)
    else:
        r = requests.get(req)
        with open(file, 'w') as result_file:
            result_file.write(r.text)
        data = json.loads(r.text)
    data_list = [[] for _ in range(len(feature))]
    time_stamp = [[] for _ in range(len(feature))]    
    stamp_bias = time.mktime((time.strptime(t+' 00:00:00','%Y-%m-%d %H:%M:%S')))
    for k in range(len(data['hourly']['data'])):
        _stamp = int((data['hourly']['data'][k]['time']-stamp_bias)/3600)
        _data = data['hourly']['data'][k]
        for m in range(len(feature)):
            try:
                data_list[m].append(_data[feature[m]])
                time_stamp[m].append(_stamp)
            except:
                pass    
    data_interp = []
    for i in range(len(feature)):
        data_interp.append(np.interp(np.arange(24), time_stamp[i], data_list[i]))
    data = np.stack(data_interp,-1)      
    return data    

def get_weather_pred(lon,lat,feature='temperature',secret_key=None):
    '''Get weather data in the future 48h. This function will have 1 darksky API use. 
    Parameters
    ----------
        lon: float, the target longtitude.
        lat: float, the target latitude.
        feature: str or list of str, the desired feature list.
        secret_key: str, a darksky secret key. if None is given, darksky_secret_key_0 in config file will be tried.
    Return
    ------
        ndarray, of shape 48*D, where D is the number of feature.
    '''
    # Parameter Check
    if not os.path.exists('data/.temp/'):
        os.mkdir('data/.temp/')
    if type(feature) is str:
        feature = [feature]
    alia_dict = {}
    alia_dict.update({x: 'temperature' for x in ['temperature','temp','tmp','wendu']})
    alia_dict.update({x: 'humidity' for x in ['humidity','hmd','hum','shidu']})
    alia_dict.update({x: 'windSpeed' for x in ['windSpeed','speed','spd','fengsu']})
    alia_dict.update({x: 'windBearing' for x in ['windBearing','direction','angel','fengxiang']})
    alia_dict.update({x: 'visibility' for x in ['visibility','kejiandu','keshidu']})
    alia_dict.update({x: 'pressure' for x in ['pressure','press','yali','qiya']}) 
    for i,x in enumerate(feature):
        if x not in alia_dict.keys():
            raise Exception(x+' is not supported.')
        feature[i] = alia_dict[x]
    if secret_key is None:
        if 'darksky_secret_key_0' in _config:
            secret_key = _config['darksky_secret_key_0']
        else:
            raise Exception('Secret key is not provide.')

    time_stamp = time.time()
    req_loc = '{0:f},{1:f}'.format(lat,lon)
    req = 'https://api.darksky.net/forecast/'+secret_key+'/'+req_loc
    file = 'data/.temp/'+req_loc+time.strftime('%Y-%m-%d-%H',time.gmtime(time_stamp+8*3600))+'.json'
    if os.path.exists(file):
        with open(file, 'r') as result_file:
            data = json.load(result_file)
    else:
        r = requests.get(req)
        with open(file, 'w') as result_file:
            result_file.write(r.text)
        data = json.loads(r.text)    
       
    data_list = [[] for _ in range(len(feature))]
    time_stamp = [[] for _ in range(len(feature))]    
    stamp_bias = data['hourly']['data'][0]['time']
    Length = len(data['hourly']['data'])
    for k in range(Length):
        _stamp = int((data['hourly']['data'][k]['time']-stamp_bias)/3600)
        _data = data['hourly']['data'][k]
        for m in range(len(feature)):
            try:
                data_list[m].append(_data[feature[m]])
                time_stamp[m].append(_stamp)
            except:
                pass    
    data_interp = []
    for i in range(len(feature)):
        data_interp.append(np.interp(np.arange(Length), time_stamp[i], data_list[i]))
    data2 = np.stack(data_interp,-1)       
    return data2[1:]


def real_time_weather(lon, lat, past_hours=81, future_hours=48, feature='temperature', secret_key=None):
    '''Get real time weather data. This function will have some darksky API use depends on how many days involved. 
    Parameters
    ----------
        lon: float, the target longtitude.
        lat: float, the target latitude.
        past_hours: int, default 81, how many hours shoud we check back from now.
        future_hours: int, default 48, how many hours shoud we go on from now, up to 48.
        feature: str or list of str, the desired feature list.
        secret_key: str, a darksky secret key. if None is given, darksky_secret_key_0 in config file will be tried.
    Return
    ------
        ndarray, of shape (past_hours+future_hours)*D, where D is the number of feature.
    To Do
    -----
        make result into a DataPack object.
    '''
    if future_hours>0:
        data_pred = get_weather_pred(lon,lat,feature,secret_key)[:future_hours]
        data_pred = [data_pred]
    elif future_hours==0:
        data_pred = []
    else:
        raise Exception('future_hours should > 0.')
    if past_hours>0:
        TH = time.gmtime(time.time()+8*3600).tm_hour + 1
        PH = (past_hours-TH)
        N = int(np.ceil(PH/24.0))
        data_past = []
        for i in range(int(N),-1,-1):
            a = get_weather_history(time.strftime('%Y-%m-%d',time.gmtime(time.time()+8*3600-i*24*3600)),lon,lat,feature,secret_key)
            data_past.append(a)
        data_past = np.concatenate(data_past,0)[N*24-PH:-(24-TH)]
        data_past = [data_past]
    elif future_hours==0:
        data_past = []
    else:
        raise Exception('past_hours should > 0.')
    data = np.concatenate(data_past+data_pred,0)
    return data



# pylint: disable=E1101
import numpy as np
import pandas as pd
import tensorflow as tf
import os
import time
import apdt
import pickle

class DataSet():
    '''DataSet can make DataPack object being Machine Learning ready.
    Description
    -----------
        Dataset object consist of three parts: pre_process, get_data, and post_process.

        - pre_process is to make DataPack obejct into a ML-ready ndarray self.tr and self.te, which have size N*D_1*D_2*...; The first dimension is considered as independent sample index; Beside this, two variable self.tr_batch_num and self.te_batch_num which claims the coressponding sample number should also be defined.
        - get_data has two key methods tr_get_batch and te_get_batch, will return a B*D_1*D_2*... as a batch of training/testing data, or list of ndarray when you need multi-input.
        - post_process defined the action that after every get_data call.

        In most case users have to re-define pre_process function to make their desired dataset structure. We also provide some pre-defined useful pre_process function for you. get_data and post_process typically don't need much work, but they can still be re-defined if you like.

    Parameters
    ----------
        datapack: DataPack object or str
            The datapack to be processed into a dataset. If a str is given, it will be considered as a pikle file path, and will construct dataset directly from that file.
        method: str
            If this argument is defined, we will use sepecific pre-defined pre_process fuction.
            Avaliable:
            - 'window': use a sliding window across time-axis to generate samples. locations in a time slice are supposed to be same (i.e. fixed-station). a samlple will be a N*T*D array. where N is the number of locations, T is window length and D is feature dimensions.
                argument for 'window':
                - split_ratio=0.7: the ratio of training set;
                - seq_len=100: the window length;
                - strides=seq_len-1: the sliding step;
                - normalize=True: if we do 0-1 normalize, Warning: dataset should always be properly normalized, and this operation is a in-place operation.
                - feature=['data0', 'data1', ...]: the feature list to be used. default to all features avaliable.

    '''
    def __init__(self, datapack, **kwarg):
        if type(datapack) is apdt.DataPack:
            self.pre_process(datapack, **kwarg)
        elif type(datapack) is str:
            self.load(datapack, **kwarg)

        self._init_counter(**kwarg)

    def dump(self, path):
        dumped_data = {'data': self.data, 'tr': self.tr, 'te': self.te}
        with open(path, 'wb') as f:
            pickle.dump(dumped_data, f, 1)

    def load(self, path, **kwarg):
        with open(path, 'rb') as f:
            dumped_data = pickle.load(f)
        self.data = dumped_data['data']
        self.tr = dumped_data['tr']
        self.te = dumped_data['te']
        self.tr_batch_num = self.tr.shape[0]
        self.te_batch_num = self.te.shape[0]

    def pre_process(self, datapack, **kwarg):
        if 'method' not in kwarg.keys():
            raise Exception("You have to re-define pre_process function or use a pre-defined one using 'method' argument.")

        if kwarg['method'] == 'window':
            if 'time-aligned' not in datapack.tag:
                raise Exception("datapack should be aligned along time axis first.")
            if 'fixed-location' not in datapack.tag:
                raise Exception("window method only support fixed station data.")
            
            if 'split_ratio' not in kwarg.keys():
                kwarg['split_ratio'] = 0.7
            if 'shuffle' not in kwarg.keys():
                kwarg['shuffle'] = True
            if 'seq_len' not in kwarg.keys():
                kwarg['seq_len'] = 100
            if 'strides' not in kwarg.keys():
                kwarg['strides'] = kwarg['seq_len'] - 1
            if 'normalize' not in kwarg.keys():
                kwarg['normalize'] = True
            if 'normalize_method' not in kwarg.keys():
                kwarg['normalize_method'] = '99pt'
            if 'feature' not in kwarg.keys():
                kwarg['feature'] = [x for x in datapack.data.columns if x.startswith('data')]

            self.method = 'window'
            T = datapack.time_length
            N = datapack.site_num
            D = len(kwarg['feature'])
            if kwarg['normalize']:
                datapack = apdt.proc.linear_normalize(datapack, kwarg['feature'], kwarg['normalize_method'])
            self.data = datapack.data.reset_index().sort_values(['datetime', 'site_id'])[kwarg['feature']].values.reshape((T, N, D))

            split_point = int(T * kwarg['split_ratio'])
            self.tr = self.data[:split_point]
            self.te = self.data[split_point:]
            self.tr_batch_num = (self.tr.shape[0]-kwarg['seq_len']+1+kwarg['strides'])//(kwarg['strides']+1)
            self.te_batch_num = (self.te.shape[0]-kwarg['seq_len']+1+kwarg['strides'])//(kwarg['strides']+1)
            if self.tr_batch_num == 0 or self.te_batch_num == 0:
                raise Exception("time_length is not enough to construct a window.")
            
            self.tr_list = []
            self.te_list = []
            for i in range(self.tr_batch_num):
                self.tr_list.append(self.tr[i*(kwarg['strides']+1):i*(kwarg['strides']+1)+kwarg['seq_len']])
            for i in range(self.te_batch_num):
                self.te_list.append(self.te[i*(kwarg['strides']+1):i*(kwarg['strides']+1)+kwarg['seq_len']])
            self.tr = np.array(self.tr_list)
            self.te = np.array(self.te_list)
            self.tr = np.transpose(self.tr, (0,2,1,3))
            self.te = np.transpose(self.te, (0,2,1,3))

    def _init_counter(self, **kwarg):
        if 'seed' in kwarg.keys():
            np.random.seed(kwarg['seed'])
        self.tr_batch_counter = 0
        self.te_batch_counter = 0
        self.tr_batch_perm = np.linspace(0, self.tr_batch_num-1, self.tr_batch_num, dtype=np.int32)
        np.random.shuffle(self.tr_batch_perm)
        self.te_batch_perm = np.linspace(0, self.te_batch_num-1, self.te_batch_num, dtype=np.int32)
        np.random.shuffle(self.te_batch_perm)

    def tr_get_batch(self, batch_size=1):
        if batch_size > self.tr_batch_num:
            batch_size = self.tr_batch_num
        if self.tr_batch_counter + batch_size > self.tr_batch_num:
            np.random.shuffle(self.tr_batch_perm)
            self.tr_batch_counter = 0
        target_index = self.tr_batch_perm[self.tr_batch_counter: self.tr_batch_counter + batch_size]
        self.tr_batch_counter = self.tr_batch_counter + batch_size
        batch = self.tr[target_index]
        batch = self.post_process(batch, mode='train')
        return batch
    
    def te_get_batch(self, batch_size=1):
        if batch_size > self.te_batch_num:
            batch_size = self.te_batch_num
        if self.te_batch_counter + batch_size > self.te_batch_num:
            np.random.shuffle(self.te_batch_perm)
            self.te_batch_counter = 0
        target_index = self.te_batch_perm[self.te_batch_counter: self.te_batch_counter + batch_size]
        self.te_batch_counter = self.te_batch_counter + batch_size
        batch = self.te[target_index]
        batch = self.post_process(batch, mode='test')
        return batch

    def post_process(self, batch, **kwarg):
        return batch

class TFModel():
    '''Provide model constructure interface based on Tensorflow (1.x).
    Description
    -----------
        There are some method of this class:

        - def_model(): you HAVE TO re-define this function to build your deep network. this fuction should define self.input, self.learning_rate, self.loss, and self.pred. self.input should be a placeholder tensor or a list of such, where training batch passed to TFModel will be sent into; self.learning_rate should be a placeholder tensor of single float value, which will control the gradient scale; self.loss should be a tensor of single float value, which will be minimized bu optimizer; self.pred should be a placeholder tensor or a list of such, which will be the output to be evaluated given some input.

        - fit(dataset): fit the model using given DataSet object. Note that please make sure the dataset structure is compatible with model's input interface.

        - eval(data): evaluate the model using given data (ndarray or a list of such, depends on self.input).

        - load(path): load parameters at given path.

    Parameters
    ----------
        - model.__init__()
            - l2_norm, float, default None.
                A L2-Regularization will apllied with weight decay parameter as l2_norm, It will be disabled when given None.
            - clip_gvs, bool, default True.
                all gradients in network will be cliiped to [-1,1] when set True.
        
        - model.fit():
            - model_name='NewModel', str.
            - lr=1e-3, float.
            - baseline=0, float. if the model loss is smaller than this, it will be automatically saved.
            - epoch=100, int. How many times should we go through the dataset.
            - print_every_n_epochs=1. int.
            - test_every_n_epochs=1. int.
            - lr_annealing='constant'. str. How to decay learning rate during training. optional: 'constant', 'step', 'cosine'.
            - lr_annealing_step_length=epoch/4. int. Avaliable when lr_annealing is set to 'step'. How often do we decay the learning rate.
            - lr_annealing_step_divisor=2.0. float. Avaliable when lr_annealing is set to 'step', we will apply lr=lr/lr_annealing_step_divisor to slow down the training process.
    '''
    def __init__(self, **kwarg):
        np.random.seed(0)
        tf.set_random_seed(0)
        self.def_model(**kwarg)
        self.setup_train_op(**kwarg)
        self.saver = tf.train.Saver()
        self.sess = tf.Session()
        self.sess.run(tf.global_variables_initializer())

    def def_model(self, **kwarg):
        pass

    def setup_train_op(self, **kwarg):
        if 'l2_norm' not in kwarg.keys():
            kwarg['l2_norm'] = None
        if 'keep_norm_loss' not in kwarg.keys():
            kwarg['keep_norm_loss'] = True
        if 'clip_gvs' not in kwarg.keys():
            kwarg['clip_gvs'] = True
        optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate)
        if kwarg['l2_norm']:
            l2_loss =  kwarg['l2_norm'] * tf.add_n([tf.nn.l2_loss(tf.cast(v, tf.float32)) for v in tf.trainable_variables()])
            if kwarg['keep_norm_loss']:
                self.loss = self.loss + l2_loss
                gvs = optimizer.compute_gradients(self.loss)
            else:
                gvs = optimizer.compute_gradients(self.loss + l2_loss)
        else:
            gvs = optimizer.compute_gradients(self.loss)
        if kwarg['clip_gvs']:
            clipped_gvs = [(tf.clip_by_value(grad, -1., 1.), var) for grad, var in gvs if grad is not None]
            gvs = [clipped_gvs[i] for i in range(0,len(clipped_gvs))]
        self.train_op = optimizer.apply_gradients(gvs)

    def load(self, path):
        self.saver.restore(self.sess, save_path=path)

    def update(self, data):
        pass

    def eval(self, data):
        if type(self.input) is list:
            feed_dict = {self.input[i]: data[i] for i in range(len(self.input))}
            feed_dict.update({self.learning_rate: 0.0})
        else:
            feed_dict = {self.learning_rate: 0.0, self.input: data}
        result = self.sess.run(self.pred, feed_dict)
        return result

    def _zip_run(self, target, feed_dict, _r=0):
        '''Recurrently unzip lists in feed_dict to generate samples.
        Description
        -----------
            Compared to self.sess.run(target, feed_dict), self._zip_run(target, feed_dict) support list values in feed_dict, where the average of unziped feed_dict will be returned.
        
        Parameters
        ----------
            target, TF tensor or a list of TF tensor
                the tensorflow tensor to be evaluated.
            feed_dict, dict
                the feed data, list values supported.
        
        Returns
        ----------
            ndarray or a multi-level list of ndarray

        Examples
        ----------
            If we run self._zip_run(self.loss, {imput:[x1,x2]}) it is equivalent to do: [self.session.run(self.loss, {input: x}) for x in [x1,x2]]
        '''
        if _r == len(feed_dict.keys()):
            return self.sess.run(target, feed_dict)
        else:
            if type(list(feed_dict.values())[_r]) is list:
                _result = []
                for i in range(len(list(feed_dict.values())[_r])):
                    _temp_dict = feed_dict.copy()
                    _temp_dict[list(feed_dict.keys())[_r]] = list(feed_dict.values())[_r][i]
                    # change _r+1 to _r to support multi-level lists.
                    _result.append(self._zip_run(target, _temp_dict, _r+1))
                return _result
            else:
                return self._zip_run(target, feed_dict, _r+1)

    def fit(self, dataset, **kwarg):
        # parameter checking
        if 'model_name' not in kwarg.keys():
            kwarg['model_name'] = 'NewModel'
        if 'lr' not in kwarg.keys():
            if 'learning_rate' in kwarg.keys():
                kwarg['lr'] = kwarg['learning_rate']
            else:
                kwarg['lr'] = 1e-3
        if 'baseline' not in kwarg.keys():
            kwarg['baseline'] = 0
        if 'epoch' not in kwarg.keys():
            kwarg['epoch'] = 100
        if 'batch_size' not in kwarg.keys():
            kwarg['batch_size'] = 1
        if 'print_every_n_epochs' not in kwarg.keys():
            kwarg['print_every_n_epochs'] = 1
        if 'test_every_n_epochs' not in kwarg.keys():
            kwarg['test_every_n_epochs'] = 1
        if 'lr_annealing' not in kwarg.keys():
            kwarg['lr_annealing'] = 'constant'
        if 'lr_annealing_step_length' not in kwarg.keys():
            kwarg['lr_annealing_step_length'] = int(kwarg['epoch']/4)
        if 'lr_annealing_step_divisor' not in kwarg.keys():
            kwarg['lr_annealing_step_divisor'] = 2.0
        if 'epoch' not in kwarg.keys():
            kwarg['epoch'] = 100

        version = time.strftime('%Y%m%d_%H%M%S')
        if not os.path.exists('model/'):
            os.mkdir('model/')
        os.mkdir('model/' + kwarg['model_name'] + version)
        lr = float(kwarg['lr'])
        performance_recorder = 1e10
        epoch_recorder = 0
        start_time = time.time()
        for epoch in range(kwarg['epoch']):
            # update an epoch
            train_ls = []
            for _ in range(dataset.tr_batch_num):
                batch = dataset.tr_get_batch(kwarg['batch_size'])
                if type(self.input) is list:
                    feed_dict = {self.input[i]: batch[i] for i in range(len(self.input))}
                    feed_dict.update({self.learning_rate: lr})
                else:
                    feed_dict = {self.learning_rate: lr, self.input: batch}
                _, ls = self.sess.run([self.train_op, self.loss], feed_dict)
                train_ls.append(ls)
            
            # print log
            if (epoch+1)%kwarg['print_every_n_epochs'] == 0:
                train_ls = np.mean(train_ls)
                print('['+kwarg['model_name']+version+']epoch ',epoch,'/',kwarg['epoch'],' Done, Train loss ',round(train_ls,4))
            
            # learning_rate_decay
            if kwarg['lr_annealing']=='step':
                if (epoch+1)%kwarg['lr_annealing_step_length'] == 0:
                        lr = lr/kwarg['lr_annealing_step_divisor']
                        print('['+kwarg['model_name']+version+']epoch ',epoch,'/',kwarg['epoch'],', Learning rate decay to ',lr)
            elif kwarg['lr_annealing']=='cosine':
                lr = float(kwarg['lr'])*(np.cos(epoch/kwarg['epoch']*np.pi)+1.0)/2

            # test
            if (epoch+1)%kwarg['test_every_n_epochs'] == 0:
                test_ls = []
                for _ in range(dataset.te_batch_num):
                    batch = dataset.te_get_batch(kwarg['batch_size'])
                    if type(self.input) is list:
                        feed_dict = {self.input[i]: batch[i] for i in range(len(self.input))}
                        feed_dict.update({self.learning_rate: lr})
                    else:
                        feed_dict = {self.learning_rate: lr, self.input: batch}
                    # ls = self.sess.run(self.loss, feed_dict)
                    ls = self._zip_run(self.loss, feed_dict)
                    test_ls.append(ls)
                test_ls = np.mean(test_ls)

                print('['+kwarg['model_name']+version+']epoch ',epoch,'/',kwarg['epoch'],' Done, Test loss ',round(test_ls,4))
                target = test_ls
                if target < performance_recorder:
                    performance_recorder = target
                    epoch_recorder = epoch
                    if target < kwarg['baseline']:
                        self.saver.save(self.sess,'model/'+kwarg['model_name']+version+'/model')
                        print('['+kwarg['model_name']+version+']epoch ',epoch,'/',kwarg['epoch'],' Model Save Success. New record ',target)
        
        print('=======Training Summary=======')
        print('Time used: ', time.strftime('%H:%M:%S',time.gmtime(time.time()-start_time)))
        if performance_recorder < 1e10:
            print('Best model at epoch ', epoch_recorder, ', with loss ',performance_recorder)
        if performance_recorder < kwarg['baseline']:
            print('Best Model saved as: ', 'model/'+kwarg['model_name']+version+'/model')
        print('Done.')            
        pass



import numpy as np
import pandas as pd
import apdt
import yaml

from apdt.config import _config

def spatial_interplot(datapack, method='NN', columns='ALL', **kwarg):
    '''Using spatial interplotion to complete the missing value.
    Parameters
    ----------
        datapack: DataPack
        method: str
            The interplotation method to be used. 
            Supported: 'NN'(Nearest Neighboor), 'IDW'(Inverse Distance Weighted)
        columns: str or list
            which data columns should we processed on. if 'ALL' is given, all data columns will be considered.
            
    Return
    ------
        DataPack
    Note
    ----
        To be supported: 'Kriging'(Kriging interplotation).
        deal with data1, data2, ...
        To Do: Why multi-thread dont work on this function?

    ''' 
    worker_num = 1
    if method=='NN':
        # A small enough scale IDW will equal to nearest neighboor.
        IDW_scale = 1e-5
    elif method=='IDW':
        try:
            IDW_scale = kwarg['IDW_scale']
        except:
            IDW_scale = 0.05
    else:
        raise Exception("Unknown interpolation method: " + method)
    if columns=='ALL':
        columns = [x for x in datapack.data.columns if x.startswith('data')]
    elif type(columns) is str:
        columns = [columns]
    
    def worker(k):
        for i, timestamp in enumerate(datapack.data.index.drop_duplicates()):
            if i%worker_num != k:
                continue
            slice_data = datapack.data.loc[timestamp].copy()
            for data_type in columns:
                nan_index = slice_data[data_type].isna()
                if not any(nan_index) or all(nan_index):
                    continue
                sloc = slice_data[~nan_index][['lon', 'lat']].values.reshape((-1,2))
                svalue = slice_data[~nan_index][data_type].values.reshape((-1,1))
                tloc = slice_data[nan_index][['lon', 'lat']].values.reshape((-1,2))
                tvalue = apdt.alz.IDW(sloc, svalue, tloc, IDW_scale)
                slice_data.loc[nan_index,data_type] = tvalue
            datapack.data.loc[timestamp] = slice_data
    
    workers = []
    for i in range(worker_num):
        workers.append(apdt.general.SubThread(worker,(i,)))
        workers[-1].start()
    for i in range(worker_num):
        workers[i].join()
    
    return datapack

def temporal_interpolate(datapack, site='ALL', method='linear'):
    '''Using temporal interplotion to complete the missing value. Note that this only works for fixed site.
    Parameters
    ----------
        datapack: DataPack
            site_info required.
        site: str or list, default 'ALL'
            the list of site to be interpolated. If given 'ALL', all available sites will be processed according to site_info.
        method: str, default 'linear'
            The interplotation method to be used. 
            Supported: 'linear'
            Reference: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.core.resample.Resampler.interpolate.html?highlight=interpolate#pandas.core.resample.Resampler.interpolate      
    Return
    ------
        DataPack
    Note
    ----
    '''
    if site=='ALL':
        site = list(datapack.site_info['site_id'])
    elif type(site) is str:
        site = [site]

    result = []
    mask = pd.Series(False, index=datapack.data.index)
    for target in site:
        _mask = datapack.data['site_id']==target
        series = datapack.data[_mask].resample(datapack.sample_unit).interpolate(method=method).bfill()
        mask = mask | _mask
        result.append(series)
    datapack.data = pd.concat([datapack.data[~mask]]+result).sort_index()
    if all(mask):
        datapack.tag.append('time-aligned')
        datapack.time_length = len(datapack.data.index.drop_duplicates())
        datapack.site_num = datapack.site_info.shape[0]
    return datapack



_config = {
  'darksky_secret_key_num': 6,
  'darksky_secret_key_0': '3636efb586ef9346635b4c2481ada25b',
  'nms_data_path': 'data/',
  'foshan_data_path': 'data/foshan_data/',
  'subthread_max_num': 4,
  }




from apdt.general import DataPack
from apdt.alz import haversine
import numpy as np

def find_sites_range(datapack, clat, clon, radius):
    '''Find sites within a circle on the map.
    Parameters
    ----------
        datapack: DataPack
            site_info is required.
        clat: float
            the lat of center point.
        clon: float
            the lon of center point.
        radius: float
            the radius of the circle. in unit km.
    Return
    ------
        list
            the list of site_id.
    '''
    if datapack.site_info is None:
        raise Exception("site_info is required.")
    loc1 = np.array([[clon, clat]])
    loc2 = datapack.site_info[['lon', 'lat']].values
    h_distance = haversine(loc1, loc2)
    h_distance[np.isnan(h_distance)] = radius + 1
    result = list(datapack.site_info['site_id'][np.squeeze(h_distance<radius)])
    return result




from apdt.general import DataPack
import numpy as np
import pandas as pd
from sklearn import gaussian_process as gp

def gp_data(time_length, site_num, dimension=1, kernel_weight=None, noise_level=None, seed=None, freq='H'):
    '''Generate a fake PM2.5-like data for quick start using gaussian process.
    Parameters
    ----------
        - time_length: int. How many hours should data have.
        - site_num: int. How many sites should data have.
        - dimension: int, default 1, How many independent sample each ST-point have.
        - kernel_weight: list of three float numbers, default [1.0, 1.0, 1.0], the relevant variance of three components: long-term trend, short-term fluction and period wave. or list of six float numbers, where addional 3 will be considered as lenth_scale parameter of three kernels(0-1, default 1).
        - noise_level: float, default 0.01, the white noise add to kernel, note that this is neccessary for long time generation.
        - seed: int. The random seed.
        - freq: str. the frequency of time stamp. 'H' for hour (default) and 'T' for minutes.
    Return
    ------
        DataPack
    Issue
    -----
    '''
    if seed is not None:
        np.random.seed(seed)
    if kernel_weight is None:
        kernel_weight = [[1.0, 1.0, 1.0]]
    if np.array(kernel_weight).ndim==1:
        kernel_weight = [kernel_weight]
    if noise_level is None:
        noise_level = 0.01
    if len(kernel_weight[0])==3:
        kernel_weight = [kernel_weight[i]+[1.0, 1.0, 1.0] for i in range(len(kernel_weight))]
    
    # Decrease this number if this program stuck.
    generation_step = 1000

    xs = np.arange(generation_step*2).reshape((generation_step*2,1))
    

    if len(kernel_weight)==1:
        # Case: Only one parameter is given, site_num will be treated as new dimension
        k1 = gp.kernels.RBF(length_scale=50.0+100.0*kernel_weight[0][3])
        k2 = gp.kernels.Matern(length_scale=20.0+10.0*kernel_weight[0][4], nu=0.5)
        k3 = gp.kernels.ExpSineSquared(length_scale=1, periodicity=100+200*kernel_weight[0][5])
        kw = gp.kernels.WhiteKernel(noise_level=noise_level)
        k = kernel_weight[0][0]*k1 + kernel_weight[0][1]*k2 + kernel_weight[0][2]*k3 + kw
        C = k(xs)
        C_11 = C[:generation_step,:generation_step]
        C_11_inv = np.linalg.inv(C_11)
        C_21 = C[generation_step:,:generation_step]
        u, s, _ = np.linalg.svd(C_11)
        us = np.matmul(u, np.diag(np.sqrt(s)))
        sample = np.matmul(us, np.random.randn(generation_step,site_num*dimension))
        C_cond = C_11 - np.matmul(np.matmul(C_21.T, C_11_inv), C_21)
        u, s, _ = np.linalg.svd(C_cond)
        us = np.matmul(u, np.diag(np.sqrt(s)))
        time_now = generation_step
        sample_list = [sample]
        while time_now < time_length:
            mu_cond = 0 + np.matmul(np.matmul(C_21, C_11_inv), sample)
            sample = np.matmul(us, np.random.randn(generation_step,site_num*dimension))
            sample = sample + mu_cond
            sample_list.append(sample)
            time_now = time_now + generation_step
        sample_list = np.concatenate(sample_list)[:time_length].reshape((time_length*site_num,dimension))
    else:
        # Case: A list of parameter is given, generate site_num samples one by one
        sample_list_all = []
        for i in range(site_num):
            k1 = gp.kernels.RBF(length_scale=50.0+100.0*kernel_weight[i][3])
            k2 = gp.kernels.Matern(length_scale=20.0+10.0*kernel_weight[i][4], nu=0.5)
            k3 = gp.kernels.ExpSineSquared(length_scale=1, periodicity=100+200*kernel_weight[i][5])
            kw = gp.kernels.WhiteKernel(noise_level=noise_level)
            k = kernel_weight[i][0]*k1 + kernel_weight[i][1]*k2 + kernel_weight[i][2]*k3 + kw
            C = k(xs)
            C_11 = C[:generation_step,:generation_step]
            C_11_inv = np.linalg.inv(C_11)
            C_21 = C[generation_step:,:generation_step]
            u, s, _ = np.linalg.svd(C_11)
            us = np.matmul(u, np.diag(np.sqrt(s)))
            sample = np.matmul(us, np.random.randn(generation_step,1*dimension))
            C_cond = C_11 - np.matmul(np.matmul(C_21.T, C_11_inv), C_21)
            u, s, _ = np.linalg.svd(C_cond)
            us = np.matmul(u, np.diag(np.sqrt(s)))
            time_now = generation_step
            sample_list = [sample]
            while time_now < time_length:
                mu_cond = 0 + np.matmul(np.matmul(C_21, C_11_inv), sample)
                sample = np.matmul(us, np.random.randn(generation_step,1*dimension))
                sample = sample + mu_cond
                sample_list.append(sample)
                time_now = time_now + generation_step
            sample_list = np.concatenate(sample_list)[:time_length].reshape((time_length*1,dimension))
            sample_list_all.append(sample_list)
        sample_list = np.stack(sample_list_all, 1).reshape((time_length*site_num, dimension))
    
    datetime_list = pd.date_range(start='2000-1-1',periods=time_length,freq=freq)
    site_name_list = ['virtual_site'+str(i) for i in range(site_num)]
    idx = pd.MultiIndex.from_product([datetime_list, site_name_list],names=('datetime', 'site_name'))
    data = pd.DataFrame(index=idx, columns=['data'+str(i) for i in range(dimension)], data=sample_list)
    data = data.reset_index()
    site_list = pd.DataFrame(data=np.random.randn(site_num, 2)+np.array([39.8673, 116.3660]),
        columns=['lat','lon'])
    site_list['site_name'] = site_name_list
    site_list['site_id'] = ['V'+str(i).zfill(4) for i in range(site_num)]
    data = data.merge(site_list, how='left', on='site_name')
    data = data.set_index('datetime')

    datapack = DataPack()
    datapack.raw_data = data
    datapack.data = data.copy()
    datapack.site_info = site_list
    datapack.data_type = ['virtual_type_'+str(i) for i in range(dimension)]
    datapack.sample_unit = freq
    datapack.tag.append('fixed-location')
    datapack.tag.append('time-aligned')
    datapack.time_length = time_length
    datapack.site_num = site_num

    return datapack




    



'''
Air Pollution Data Toolbox
--------------------------

Provide useful tools for air pollution data research.

Check https://github.com/Zhiyuan-Wu/apdt for more information.
'''
from apdt.general import DataPack, SubThread
from apdt import io
from apdt import alz
from apdt import proc
from apdt import ml


# -*- coding: UTF-8 -*-
from apdt.general import DataPack
import numpy as np
import pandas as pd
import yaml
import time
import json
import os

from apdt.config import _config

def foshan_data(site, start_date, end_date=None, gas='pm2d5'):
    '''Load fixed sensor data collected from foshan.
    Parameters
    ----------
        site: string or list
            The station name of queried stations like "体育馆岗亭", or list of such. Sepcifically, if given 'ALL', all available sites will be used.
        start_date: string
            The start time (include) like "2018-11-01".
        end_date: string, default: same as start_date
            The end time (include) like "2018-11-30"
        gas: string or list, default: 'pm2d5'
            The gas type to be collected.
    Return
    ------
        DataPack
            The loaded data that can be used by apdt.
    '''
    # Parameter Check
    data_path = _config['foshan_data_path']
    air = gas # To Do: support a list of gas type.

    if site=='ALL':
        site = sorted(os.listdir(data_path))
    elif type(site) is str:
        site = [site]
    site = [x if x.endswith('.xls') else x+'.xls' for x in site]

    if start_date < '2018-11-01' or start_date > '2018-11-30':
        raise Exception("Foshan data range between '2018-11-01'-'2018-11-30'")
    if end_date is None:
        end_date = start_date
    if end_date < '2018-11-01' or end_date > '2018-11-30':
        raise Exception("Foshan data range between '2018-11-01'-'2018-11-30'")
    if end_date < start_date:
        raise Exception('end_date shoule larger than start_date.')
    
    datalist = []
    location = []
    
    for id,target in enumerate(site):
        data = pd.read_excel(data_path+target, index_col=0)
        dat = data[[air]+['lon', 'lat']].dropna()
        dat = dat.reset_index().rename(columns={'index':'datetime',air:'data0', 'lon':'lon', 'lat':'lat'}).set_index('datetime')
        dat = dat.resample("T").asfreq()
        dat = dat[dat.index>=start_date + ' 00:00:00']
        dat = dat[dat.index<=end_date + ' 23:59:59']
        # nanidx=dat.index[np.isnan(dat[air])]
        dat=dat.dropna()
        dat=dat.resample("T").bfill()
        # dat["isnan"]=0
        # dat["isnan"].loc[nanidx]=1
        dat['site_id'] = 'foshan' + str(id)
        dat['site_name'] = target.split('.')[0]
        dat["date"]=dat.index.map(lambda x:x.date())
        
        datalist.append(dat)
        location.append(['foshan' + str(id), target.split('.')[0], dat['lon'][0], dat['lat'][0]])

    datalist = pd.concat(datalist)
    location = pd.DataFrame(location, columns=['site_id', 'site_name', 'lon', 'lat'])
    artifact = DataPack()
    artifact.raw_data = datalist
    artifact.data = datalist.copy()
    artifact.site_info = location
    artifact.data_type = gas
    artifact.sample_unit = 'M'
    artifact.tag.append('fixed-location')
    artifact.tag.append('time-aligned')
    return artifact
    


"""Some general definations of apdt.
"""
import numpy as np
import pandas as pd
import threading

class DataPack():
    """DataPack is the standard data structure used by apdt.
    """
    def __init__(self):
        _empty_data = pd.DataFrame(columns=['datetime', 'lon', 'lat', 'data0'])
        _empty_data['datetime'] = pd.to_datetime(_empty_data['datetime'])
        _empty_data = _empty_data.set_index('datetime')
        self.raw_data = _empty_data
        self.data = _empty_data.copy()
        self.site_info = None
        self.data_type = ['data0']
        self.sample_unit = None
        self.tag = []

    def reset(self):
        '''Reset data into just-loaded status.
        '''
        self.data = self.raw_data.copy()

    def concat(self, other):
        '''combine two datapack into one.
        To Do
        -----
            - self.tag should be properly processed.
            - Non-inplace version?
        '''
        self.data.rename(columns={'data' + str(i): j for i,j in enumerate(self.data_type)}, inplace=True)
        target_data = other.data.rename(columns={'data' + str(i): j for i,j in enumerate(other.data_type)}, inplace=False)
        self.data = pd.concat([self.data, target_data], sort=True)
        self.data_type = list({}.fromkeys(self.data_type + other.data_type).keys())
        self.data.rename(columns={j: 'data' + str(i) for i,j in enumerate(self.data_type)}, inplace=True)
        
        if other.site_info is not None:
            if self.site_info is None:
                self.site_info = other.site_info.copy()
            else:
                self.site_info = pd.concat([self.site_info, other.site_info]).drop_duplicates()

        if self.site_info is not None:
            self.site_num = self.site_info.shape[0]

        sample_unit_list = ['S', 'M', 'H', 'D', 'W', 'M', 'Y']
        self.sample_unit = sample_unit_list[min(sample_unit_list.index(self.sample_unit), 
                sample_unit_list.index(other.sample_unit))]
        return self
    
    def merge(self, other, on=['datetime','lat','lon']):
        '''Merge two datapack into one, or concat two along dimension 1.
        '''
        for x in self.data_type:
            if x in other.data_type:
                raise Exception('Merge error: two datapack have same data type ' + x)
        self.data.rename(columns={'data' + str(i): j for i,j in enumerate(self.data_type)}, inplace=True)
        target_data = other.data.rename(columns={'data' + str(i): j for i,j in enumerate(other.data_type)}, inplace=False)
        self.data = self.data.reset_index()
        target_data = target_data.reset_index()
        target_data = target_data[on+other.data_type]
        self.data = pd.merge(self.data, target_data, how='left',on=['datetime','lat','lon'])
        self.data_type = list({}.fromkeys(self.data_type + other.data_type).keys())
        self.data.rename(columns={j: 'data' + str(i) for i,j in enumerate(self.data_type)}, inplace=True)
        self.data = self.data.set_index('datetime')
        return self




    def add_row(self, datetime, lat=None, lon=None, value=None):
        '''Append rows into datapack. This fuction is usually used as virtual-sampling.
        '''
        pass
    
    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__repr__()

    def __add__(self, other):
        return self.concat(other)

class SubThread(threading.Thread):
    """SubThread is an interface of python threading.
    """
    def __init__(self, func, args = ()):
        super(SubThread, self).__init__()
        self.func = func
        self.args = args
    
    def run(self):
        self.result = self.func(*self.args)

    def get_result(self):
        try:
            return self.result
        except Exception:
            return None





# -*- coding: UTF-8 -*-
from apdt.general import DataPack, SubThread
import numpy as np
import pandas as pd
import yaml
import time
import json

from apdt.config import _config

def load_nms(site, start_date, end_date=None, gas='pm2d5'):
    '''Load national monitoring stations (NMS) data.
    Parameters
    ----------
        site: string or list
            The station ID of queried stations like "1001A", or list of such. Sepcifically, if given 'ALL', all available sites will be used.
        start_date: string
            The start time (include) like "2017-01-01".
        end_date: string, default: same as start_date
            The end time (include) like "2018-12-31"
        gas: string or list, default: 'pm2d5'
            The gas type to be collected.
    Return
    ------
        DataPack
            The loaded data that can be used by apdt.
    '''
    worker_num = _config['subthread_max_num']
    data_path = _config['nms_data_path']
    location = pd.read_csv(data_path+'site_location.csv')

    if site=='ALL':
        site = list(location['监测点编码'])
    elif type(site) is str:
        site = [site]
    gas = {'pm2d5': 'PM2.5', 'pm10': 'PM10', 'aqi': 'AQI', 'so2': 'SO2', 'no2': 'NO2',\
        'o3': 'O3', 'co': 'CO', 'co2': 'CO2'}[gas]
    if end_date is None:
        end_date = start_date

    
    location = location[[location.iloc[i,0] in site for i in range(location.shape[0])]]
    location['监测点名称'] = location['城市'] + '-' + location['监测点名称']
    location = location.rename(columns={'监测点编码': 'site_id', '监测点名称': 'site_name', '经度': 'lon', '纬度': 'lat'})
    location = location[['site_id', 'site_name', 'lat', 'lon']]

    
    #Collect from file
    start_time = time.mktime((time.strptime(start_date+' 12:00:00', '%Y-%m-%d %H:%M:%S')))
    end_time = time.mktime((time.strptime(end_date+' 12:00:00', '%Y-%m-%d %H:%M:%S')))
    template = pd.DataFrame({'hour': [i for i in range(24)]})

    def worker(time_now, end_time):
        data_list = []
        while time_now < end_time:
            file = data_path + 'site' + \
                    str(time.gmtime(time_now).tm_year) + \
                    '/china_sites_' + \
                    time.strftime('%Y%m%d', time.gmtime(time_now)) + '.csv'
            try:
                data = pd.read_csv(file)
                data = data.reindex(columns=list(set(site).union(set(data.columns))))
                data = data.loc[data['type']==gas][['hour'] + site]
                data = template.merge(data, on='hour', how='left')
                data['hour'] = time.strftime('%Y%m%d', time.gmtime(time_now)) + ' ' + data['hour'].astype(str) + ':00:00'
                data['hour'] = pd.to_datetime(data['hour'])
                data = data.rename(columns={'hour': 'datetime'}).set_index('datetime')
                data = data.stack(dropna=False).reset_index(level = 1, drop = False).rename(columns={'level_1': 'site_id', 0: 'data0'})
                data_list.append(data)
            except:
                pass
            time_now += 24*3600
        return data_list
    workers = []
    data_list = []
    time_list = [start_time] + [int((end_time-start_time)/3600/24/worker_num)*k*24*3600+start_time for k in range(1,worker_num)] + [end_time+1]
    for i in range(worker_num):
        workers.append(SubThread(worker,(time_list[i],time_list[i+1])))
        workers[-1].start()
    for i in range(worker_num):
        workers[i].join()
        data_list = data_list+workers[i].get_result()
    data_list = pd.concat(data_list)
    data_list = data_list.reset_index().merge(location, on='site_id', how='left').set_index('datetime')
    data_list = data_list.dropna(subset=['lat', 'lon'])
    location = location.dropna(subset=['lat', 'lon'])

    artifact = DataPack()
    artifact.raw_data = data_list
    artifact.data = data_list.copy()
    artifact.site_info = location
    artifact.data_type = [gas]
    artifact.sample_unit = 'H'
    artifact.tag.append('fixed-location')

    return artifact


def load_weather(site, start_date, end_date=None, feature='temperature', **kwarg):
    '''Load national monitoring stations (NMS) weather data.
    Parameters
    ----------
        site: string or list
            The station ID of queried stations like "1001A", or list of such.
        start_date: string
            The start time (include) like "2017-01-01".
        end_date: string, default: same as start_date
            The end time (include) like "2018-12-31"
        feature: string or list, default: 'temperature'
            The data type to be collected.
    Return
    ------
        DataPack
            The loaded data that can be used by apdt.
    To Do
    -----
        Deal with alias.
        Deal with unit transform.
        Deal with data download.
        Deal with coordinates input.
        Deal with multi-thread
        Deal with site='ALL'
    '''

    if 'enable_warning' not in kwarg.keys():
        kwarg['enable_warning'] = True

    data_path = _config['nms_data_path']
    location = pd.read_csv(data_path+'site_location.csv')

    if site=='ALL':
        # site = list(location['监测点编码'])
        print('all site not supported yet')
        return None
    elif type(site) is str:
        site = [site]
    if end_date is None:
        end_date = start_date
    if type(feature) is str:
        feature = [feature]
    
    alia_dict = {}
    alia_dict.update({x: 'temperature' for x in ['temperature','temp','tmp','wendu']})
    alia_dict.update({x: 'humidity' for x in ['humidity','hmd','hum','shidu']})
    alia_dict.update({x: 'windSpeed' for x in ['windSpeed','speed','spd','fengsu']})
    alia_dict.update({x: 'windBearing' for x in ['windBearing','direction','angel','fengxiang']})
    alia_dict.update({x: 'visibility' for x in ['visibility','kejiandu','keshidu']})
    alia_dict.update({x: 'pressure' for x in ['pressure','press','yali','qiya']}) 
    for i,x in enumerate(feature):
        if x not in alia_dict.keys():
            raise Exception(x+' is not supported.')
        feature[i] = alia_dict[x]
        

    
    location = location[[location.iloc[i,0] in site for i in range(location.shape[0])]]
    location['监测点名称'] = location['城市'] + '-' + location['监测点名称']
    location = location.rename(columns={'监测点编码': 'site_id', '监测点名称': 'site_name', '经度': 'lon', '纬度': 'lat'})
    location = location[['site_id', 'site_name', 'lat', 'lon']]

    data_bag = []
    for site_id in site:
        #Collect from file
        start_time = time.mktime((time.strptime(start_date+' 12:00:00','%Y-%m-%d %H:%M:%S')))
        end_time = time.mktime((time.strptime(end_date+' 12:00:00','%Y-%m-%d %H:%M:%S')))
        time_now = start_time
        stamp_bias = time.mktime((time.strptime(start_date+' 00:00:00','%Y-%m-%d %H:%M:%S')))
        
        data_list = [[] for _ in range(len(feature))]
        time_stamp = [[] for _ in range(len(feature))]
        while time_now<=end_time:
            file = 'data/weather/'+site_id+'/'+time.strftime('%Y-%m-%d',time.gmtime(time_now))+'.json'
            with open(file) as f:
                data = json.load(f)
                if 'hourly' in data.keys():
                    for k in range(len(data['hourly']['data'])):
                        _stamp = int((data['hourly']['data'][k]['time']-stamp_bias)/3600)
                        _data = data['hourly']['data'][k]
                        for m in range(len(feature)):
                            try:
                                data_list[m].append(_data[feature[m]])
                                time_stamp[m].append(_stamp)
                            except:
                                pass
            time_now += 24*3600

        # Linear interploation
        Length = int((end_time-start_time)//3600 + 24)
        data_ratio = np.array([len(data_list[m]) for m in range(len(feature))])/Length
        if np.any(data_ratio < 0.2) and kwarg['enable_warning']:
            print(__file__+' Runtime Warning: ')
            print('Too much missing value in site ' + site_id + ' Weather Data, the returned interpolated data may be not meaningful.')
        data_interp = []
        for i in range(len(feature)):
            data_interp.append(np.interp(np.arange(Length), time_stamp[i], data_list[i]))
        data = np.stack(data_interp,-1)

        data_table = pd.DataFrame(data=data,columns=['data'+str(i) for i in range(len(feature))])
        data_table['site_id'] = site_id
        data_table['datetime'] = pd.date_range(start=start_date,periods=Length,freq='H')
        data_bag.append(data_table)

    data_bag = pd.concat(data_bag)
    data_bag = data_bag.merge(location, on='site_id', how='left').set_index('datetime')
    data_bag = data_bag.dropna(subset=['lat', 'lon'])
    location = location.dropna(subset=['lat', 'lon'])

    artifact = DataPack()
    artifact.raw_data = data_bag
    artifact.data = data_bag.copy()
    artifact.site_info = location
    artifact.data_type = feature
    artifact.sample_unit = 'H'
    artifact.tag.append('fixed-location')
    artifact.tag.append('time-aligned')

    return artifact


from setuptools import setup
import setuptools
with open("README.md") as fh:
    long_description = fh.read()
    
if __name__ == '__main__':
    setup(name = 'apdt',
          version = '0.1',
          description = 'apdt',
          author = 'Zhiyuan-Wu',
          author_email = '86562713@qq.com',
          url = 'https://github.com/Zhiyuan-Wu/apdt',
          maintainer = 'Zhiyuan-Wu',
          maintainer_email = '86562713@qq.com',
          long_description = long_description,
          long_description_content_type="text/markdown",          
          install_requires = ['numpy', 'pandas', 'sklearn'],
          license = 'Apache License Version 2.0',
          packages=setuptools.find_packages(),
          include_package_data=True,
          classifiers = (
              "Development Status :: 4 - Beta",
              "Programming Language :: Python :: 3.7",
              "Programming Language :: Python :: 3.6",
              "Operating System :: OS Independent",
          ),
          )



from apdt.io.loadnms import load_nms, load_weather
from apdt.io.fake_data import gp_data
from apdt.io.realtime import get_weather_history, get_weather_pred, real_time_weather
from apdt.io.loadfoshan import foshan_data


# pylint: disable=E1101
import tensorflow as tf
import numpy as np

def mlp_weight(**kwarg):
    '''Construct weight for MLP.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
        input_dim: int, default 10
            the number of input neurals
        output_dim: int, default 10
            the number of output neurals
        n_hidden: list of int, default [128,128]
            the number of hidden units, the length of list decide the number of layers.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'mlp'
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 10
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 10
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = [128, 128]
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True
    
    w = {}
    with tf.variable_scope(kwarg['name']):
        hn = kwarg['input_dim']
        for i in range(len(kwarg['n_hidden'])):
            w['weight'+str(i)] = tf.get_variable('weight'+str(i),[hn, kwarg['n_hidden'][i]],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
            w['bias'+str(i)] = tf.get_variable('bias'+str(i),[kwarg['n_hidden'][i],],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
            hn = kwarg['n_hidden'][i]
        w['weightout'] = tf.get_variable('weightout',[hn, kwarg['output_dim']],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
        w['biasout'] = tf.get_variable('biasout',[kwarg['output_dim'],],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
    
    return w

def MLP(input, n_layers=None, output_dim=None, n_hidden=None, weights=None, name='mlp', **kwarg):
    '''Map a vector into another using MLP.

    parameter set (n_layers, output_dim, n_hidden) will be inferenced from weights if it exists. However, in the case you have modified weights from mlp_weight (e.g. combine two weights together) things may comes wrong. Please also give (n_layers, output_dim, n_hidden).

    Parameter
    ---------
        input: tensor
            a tensor shaped [batch_size, feature_dim]
        n_layers: int
            the number of hidden layers.
        output_dim: int
            the dimension of output vector.
        n_hidden: list of int
            the number of hidden neural in each layer.
        weights: dict
            a dictionary of weights generated by apdt.ml.mlp_weight
    Return
    ------
        tensor
            The output vector, shaped [batch_size, feature_dim_output]
    Note
    ----
        Dropout is still not available now, because it shoule support automatic switch with training/testing condition.
        BN is not supported now.
    '''
    
    if weights is None:
        if n_layers is None:
            n_layers = 2
        if output_dim is None:
            output_dim = 10
        if n_hidden is None:
            n_hidden = [128]*n_layers
        weights = mlp_weight(name=name,input_dim=input.shape[-1].value,output_dim=output_dim,n_hidden=n_hidden)
    else:
        if n_layers is None:
            n_layers = len(weights.keys())//2 - 1
        if output_dim is None:
            output_dim = weights['biasout'].shape[0].value

    if 'activation' not in kwarg:
        kwarg['activation'] = tf.nn.relu
    if 'dropout' not in kwarg:
        kwarg['dropout'] = False
    if 'dropout_ratio' not in kwarg:
        kwarg['dropout_ratio'] = 0.5
    
    x = input
    x_shape = x.shape
    x = tf.reshape(x,[-1,x_shape[-1].value])
    for i in range(n_layers):
        x = tf.matmul(x, weights['weight'+str(i)]) + weights['bias'+str(i)]
        x = kwarg['activation'](x)
        if kwarg['dropout']:
            x = tf.nn.dropout(x, kwarg['dropout_ratio'])
    x = tf.matmul(x, weights['weightout']) + weights['biasout']
    x = tf.reshape(x,list(x_shape)[:-1]+[tf.Dimension(output_dim)])
    return x
    
    

def wavenet_weight(**kwarg):
    '''Construct weight for wavenet.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'wavenet'
    if 'res_channel' not in kwarg:
        kwarg['res_channel'] = 32
    if 'skip_channel' not in kwarg:
        kwarg['skip_channel'] = 16
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 1
    if 'DilatedConvLayers' not in kwarg:
        kwarg['DilatedConvLayers'] = 4
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 128
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True
    if 'share_param_between_layers' not in kwarg:
        kwarg['share_param_between_layers'] = False

    with tf.variable_scope(kwarg['name']):
        res_channel=kwarg['res_channel']
        skip_channel=kwarg['skip_channel']
        bits = 5
        w = {}
        if kwarg['share_param_between_layers']:
            k = 0
            w['kernel_l'+str(k)] = tf.get_variable('kernel_l'+str(k),[3, res_channel, 2*res_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
            w['1_by_1_res_l'+str(k)] = tf.get_variable('1_by_1_res_l'+str(k),[1, res_channel, res_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
            w['1_by_1_skip_l'+str(k)] = tf.get_variable('1_by_1_skip_l'+str(k),[1, res_channel, skip_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
            for k in range(1, kwarg['DilatedConvLayers']):
                w['kernel_l'+str(k)] = w['kernel_l'+str(0)]
                w['1_by_1_res_l'+str(k)] = w['1_by_1_res_l'+str(0)]
                w['1_by_1_skip_l'+str(k)] = w['1_by_1_skip_l'+str(0)]
        else:
            for k in range(0, kwarg['DilatedConvLayers']):
                w['kernel_l'+str(k)] = tf.get_variable('kernel_l'+str(k),[3, res_channel, 2*res_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
                w['1_by_1_res_l'+str(k)] = tf.get_variable('1_by_1_res_l'+str(k),[1, res_channel, res_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
                w['1_by_1_skip_l'+str(k)] = tf.get_variable('1_by_1_skip_l'+str(k),[1, res_channel, skip_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
        w['1_by_1_x'] = tf.get_variable('1_by_1_x', [1, kwarg['input_dim'], res_channel],initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
        w['1_by_1_skip1'] = tf.get_variable(
                    '1_by_1_skip1', [1, skip_channel*kwarg['DilatedConvLayers']+kwarg['input_dim'], kwarg['n_hidden']],
                    initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
        w['1_by_1_skip2'] = tf.get_variable(
                    '1_by_1_skip2', [1, kwarg['n_hidden'], 2**bits],
                    initializer=tf.contrib.layers.xavier_initializer(), trainable=kwarg['trainable'])
        return w

def WaveNet(input, weights, name, **kwarg):
    '''Map a time series into another using WaveNet.
    Parameter
    ---------
        input: tensor
            a tensor shaped [batch_size, time_length, feature_dim]
        weights: dict
            a dictionary of weights generated by apdt.ml.wavenet_weight
        name:
            the variable name scope to be used.
    Return
    ------
        tensor
            The corresponding prediction series. a tensor shaped [batch_size, time_length, 1]
        tensor
            The loss function.
    '''
    if 'res_channel' not in kwarg.keys():
        kwarg['res_channel'] = 32
    if 'skip_channel' not in kwarg.keys():
        kwarg['skip_channel'] = 16
    if 'input_dim' not in kwarg.keys():
        kwarg['input_dim'] = 1
    if 'DilatedConvLayers' not in kwarg.keys():
        kwarg['DilatedConvLayers'] = 4
    if 'n_hidden' not in kwarg.keys():
        kwarg['n_hidden'] = 128
    if 'dilated' not in kwarg.keys():
        kwarg['dilated'] = 3
    if 'loss' not in kwarg.keys():
        kwarg['loss'] = 'MAE'
    if 'bits' not in kwarg.keys():
        kwarg['bits'] = 5
    if 'return_type' not in kwarg.keys():
        kwarg['return_type'] = 'pred+loss'

    def CasualResUnit(x,k,res_channel=kwarg['res_channel'],skip_channel=kwarg['skip_channel'],name='CasualResUnit'):
        k = int(np.log(k)/np.log(kwarg['dilated']))
        w = weights['kernel_l'+str(k)]
        _x = tf.pad(x,[[0,0],[2,0],[0,0]])
        y = tf.nn.conv1d(_x, w, 1, 'VALID')
        h,g = tf.split(y, 2, axis=-1)
        h = tf.tanh(h)
        g = tf.sigmoid(g)
        h = h*g
        w = weights['1_by_1_res_l'+str(k)]
        o = tf.nn.conv1d(h, w, 1, 'SAME')
        o = o+x
        w = weights['1_by_1_skip_l'+str(k)]
        skip = tf.nn.conv1d(h, w, 1, 'SAME')
        return o,skip
    
    def DilatedConv(x,dilated=kwarg['dilated'],res_channel=kwarg['res_channel'],skip_channel=kwarg['skip_channel'],name='DilatedConv'):
        n = x.shape[-1].value
        l = x.shape[1].value
        if l%dilated != 0:
            print('Dilated Error, when dilated at '+str(dilated))
            exit()
        num = l//dilated
        with tf.variable_scope(name) as scope:
            x = tf.reshape(x,[-1,num,dilated,n])
            _out = []
            _skip = []
            for i in range(dilated):
                out, skip = CasualResUnit(x[:,:,i],dilated,res_channel=res_channel,skip_channel=skip_channel,name='CasualResUnit#'+str(i))
                _out.append(out)
                _skip.append(skip)
                if i==0:
                    scope.reuse_variables()
            o = tf.stack(_out,axis=2)
            o = tf.reshape(o,[-1,l,n])
            skip = tf.stack(_skip,axis=2)
            skip = tf.reshape(skip,[-1,l,skip_channel])
            return o,skip
    
    def u_law_encoder(x,bits=5):
        bits = 2**bits
        x = tf.clip_by_value(x,0.0,0.99999)
        x = tf.floor(tf.log(1+bits*x)*bits/np.log(1+bits))
        x = tf.one_hot(tf.cast(tf.squeeze(x,-1),tf.int32),depth=bits)
        return x
    
    def u_law_decoder(x,bits=5):
        x = tf.cast(x,tf.float32)
        x = (tf.exp((x+0.5)/(2**bits)*np.log(1.0+(2**bits)))-1.0)/(2**bits)
        return tf.expand_dims(x,-1)
    
    def SingleChannelNetwork(x,channel,bits,encoder,decoder,name='Channel'):
        with tf.variable_scope(name):
            y = x[:,1:,channel:channel+1]
            x = x[:,:-1,:]
            _skip = [x]
            w = weights['1_by_1_x']
            x = tf.nn.conv1d(x, w, 1, 'SAME')
            for i in range(kwarg['DilatedConvLayers']):
                if i==0:
                    o,skip = DilatedConv(x,dilated=3**i,name='DilatedConv'+str(i))
                    _skip.append(skip)
                else:
                    o,skip = DilatedConv(o,dilated=3**i,name='DilatedConv'+str(i))
                    _skip.append(skip)
            skip = tf.concat(_skip,axis=-1)
            skip = tf.nn.relu(skip)
            skip_raw = skip
            w = weights['1_by_1_skip1']
            skip = tf.nn.conv1d(skip, w, 1, 'SAME')
            skip = tf.nn.relu(skip)
            w = weights['1_by_1_skip2']
            pred = tf.nn.conv1d(skip, w, 1, 'SAME')
            pred_prob = tf.nn.softmax(pred)
            pred = tf.nn.conv1d(pred_prob, decoder(tf.constant(np.arange(2**bits).reshape(1,2**bits))), 1, 'SAME')
            if kwarg['loss']=='MAE':
                loss = tf.reduce_mean(tf.abs(pred-y))
            elif kwarg['loss']=='RMSE':
                loss = tf.sqrt(tf.reduce_mean(tf.square(pred-y)))
            return pred,loss,skip_raw
    
    pred,loss,skip_raw = SingleChannelNetwork(input,0,kwarg['bits'],u_law_encoder,u_law_decoder,name=name)
    if kwarg['return_type'] == 'pred+loss':
        return pred,loss
    elif kwarg['return_type'] == 'feature':
        return skip_raw
pass


from apdt.general import DataPack
import numpy as np
import pywt
import pandas as pd

def groups(dat,type):
    dgroups=dat.groupby(dat[type])
    return dgroups 

def group_info(dgroup,air='None'):
    date=dgroup[0]
    dg=dgroup[1]
    if air=='None':
        return date,dg
    else:
        a=dg[air]
        return date,dg,a
    
def re_mean(array_like):
    try:
        np.nan_to_num(array_like)
        v80=np.percentile(array_like,75)
        v20=np.percentile(array_like,25)
        v_array=array_like[array_like>=v20]
        v_array=v_array[v_array<=v80]
        return np.mean(v_array)
    except IndexError:
        return np.nan

def re_std(array_like):
    try:
        np.nan_to_num(array_like)
        v80=np.percentile(array_like,75)
        v20=np.percentile(array_like,25)
        v_array=array_like[array_like>=v20]
        v_array=v_array[v_array<=v80]
        return np.std(v_array)
    except IndexError:
        return np.nan
        
def VIX_ln(array):
    logarray=np.log(array)
    return np.std(logarray) 

def wavelet_decomp(data, step, w="haar"):
    w = pywt.Wavelet(w)
    a = data
    ca = []
    cd = []
    for _ in range(step):    
        (a, d) = pywt.dwt(a, w)
        ca.append(a)
        cd.append(d)
    return ca,cd

def wavelet_comp(ca, cd, w="haar"):
    rec_a = []
    rec_d = []

    for i, coeff in enumerate(ca):
        coeff_list = [coeff, None] + [None] * i
        rec_a.append(pywt.waverec(coeff_list, w))

    for i, coeff in enumerate(cd):
        coeff_list = [None, coeff] + [None] * i
        rec_d.append(pywt.waverec(coeff_list, w))

    return rec_a,rec_d

def get_theta(dg, data_type):
    dg_mean = re_mean(dg[data_type]) # the mean of 25~75 percentage
    dg_vix = VIX_ln(dg[data_type]) # std(log(x))
    dg['daymean'] = dg_mean
    dg['dayVIX'] = dg_vix
    theta = dg_mean * dg_vix
    return theta

def threshold(d,theta):
    for dd in d:
        # dd_mean=re_mean(dd)
        # whats this?????
        dd[dd>theta]=0
        dd[dd<theta]=0


def after_filter(dg,ra,rd,lv=5):
    rdd=[]
    for dd in rd:
        rdd.append(dd[0:len(dg)])
    x=np.array(rdd[0:lv])
    dx=np.sum(x,axis=0)
    ax=ra[lv-1]
    dg["after_filter"]=ax[0:len(dg)]+dx
    return dg

def smooth(dat,affiltdata,dev):
    dat["after_filter"]=affiltdata["after_filter"].rolling(60,min_periods=1,center=True).mean()
    dat["devid"]=dev
    dat["day_mean"]=affiltdata["daymean"]
    dat["day_VIX"]=affiltdata["dayVIX"]
    

def short_event(dat, data_type):
    dat["thresh"]=dat["after_filter"]+dat["day_VIX"]*3
    singleshortdata=dat[dat[data_type]-dat["thresh"]>0]
    return singleshortdata

def longterm_data(dat,dev,level=4,theta=30):
    longdata=pd.DataFrame()
    longtermdat=dat.resample("10T").mean()
    longtermdat["air_10"]=longtermdat['after_filter']
    longdata[dev]=longtermdat["air_10"]
    longtermdat=longtermdat[["air_10","day_mean","day_VIX"]]
    a,d=wavelet_decomp(longtermdat["air_10"],level)
    threshold(d,theta)
    ra,rd=wavelet_comp(a,d)
    rdd=[]
    for dd in rd:
        rdd.append(dd[0:len(longtermdat)])
        
    x=np.array(rdd[0:level])
    dx=np.sum(x,axis=0)
    ax=ra[level-1]
    longtermdat["after_filter"]=ax[0:len(longtermdat)]+dx
    longtermdat["after_filter"]=longtermdat["after_filter"].rolling(18,min_periods=1,center=True).mean()
    longtermdat["thresh"]=longtermdat["after_filter"]+(longtermdat["day_VIX"]*3)
    return longtermdat

def long_event(data, site_name):
    temp=pd.DataFrame()
    simglelongevent=data[data["air_10"]-data["thresh"]>0]
    simglelongevent["devid"]=site_name
    temp=temp.append(simglelongevent)
    return temp


def get_df_name(df):
    name =[x for x in globals() if globals()[x] is df][0]
    return name

def save_csv(path,data):
    data["time"]=data.index
    name=get_df_name(data)
    data.to_csv(path+name+'.csv',index=False)

def small_scale(dorg,data_type='data0',out=False,outpath=None):
    
    # dorg=pd.read_csv(datapack)
    #dat.index=pd.to_datetime(dat.time)
    dorg['time'] = dorg.index
    dgs=dorg.groupby(dorg["devid"])
    allevent=pd.DataFrame()
    for dg in dgs:
        dev=dg[0]
        data=dg[1]
    
        data.loc[:,"time"]=pd.to_datetime(data["time"])
        data=data.sort_values(by=['time'])
        data["tshift"]=data["time"].shift()
        data["tdiff"]=data["time"]-data["tshift"]
        data=data.reset_index()
        
        df = data[data.tdiff > pd.Timedelta("2m")]
        start_idx = [0]
        start_idx.extend(df.index)
        end_idx = list(df.index - 1)
        end_idx.append(len(data) - 1)
        end_time = data.iloc[end_idx]['time'].reset_index(drop=True)
        start_time = data.iloc[start_idx]['time'].reset_index(drop=True)
        con_time=end_time-start_time
        
        env_strength = []
        for i in range(len(start_idx)):
            d = data.iloc[start_idx[i]:end_idx[i]]
            if len(d)>0:
                env_strength.append(np.mean(d[data_type]-d["after_filter"]))
            else:
                env_strength.append(np.nan)
        
        dat=pd.DataFrame()
        dat["starttime"] = data.loc[start_idx]["time"].reset_index(drop=True)
        dat["timedelta"]=con_time.map(lambda x:x.seconds/60)
        
        dat["start_value"] = data.loc[start_idx][data_type].reset_index(drop=True)  
        dat["single_strength"] = env_strength
    
        dat["month"]=dat["starttime"].map(lambda x:x.month)
        dat["weekday"]=dat["starttime"].map(lambda x:x.weekday())
        dat["hour"]=dat["starttime"].map(lambda x:x.hour)
        dat["dev"] = dev  
        allevent=allevent.append(dat)
	
    # sevent=allevent.dropna()
    if out==True:
        if outpath==None:
            allevent.to_csv("gathershortevent.csv")
        else:
            allevent.to_csv(outpath+"gathershortevent.csv")
    return allevent

def big_scale(dorg,time_delta=10,strength=10,out=False,outpath=None):
    
    #dorg=pd.read_csv(datapack)
    #dat.index=pd.to_datetime(dat.time)
    dorg['time'] = dorg.index
    dgs=dorg.groupby(dorg["devid"])
    allevent=pd.DataFrame()
    for dg in dgs:
        dev=dg[0]
        data=dg[1]
        
        data.loc[:,"time"]=pd.to_datetime(data["time"])
        data=data.sort_values(by=['time'])
        data["tshift"]=data["time"].shift()
        data["tdiff"]=data["time"]-data["tshift"]
        data=data.reset_index()
        
        df = data[data.tdiff > pd.Timedelta("20m")]
        start_idx = [0]
        start_idx.extend(df.index)
        end_idx = list(df.index - 1)
        end_idx.append(len(data) - 1)
        end_time = data.iloc[end_idx]['time'].reset_index(drop=True)
        start_time = data.iloc[start_idx]['time'].reset_index(drop=True)
        con_time=end_time-start_time
        
        env_strength = []
        for i in range(len(start_idx)):
            d = data.iloc[start_idx[i]:end_idx[i]]
            if len(d)>0:
                env_strength.append(np.mean(d["air_10"]-d["after_filter"]))
            else:
                env_strength.append(np.nan)
        
        dat=pd.DataFrame()
        dat["starttime"] = data.loc[start_idx]["time"].reset_index(drop=True)
        dat["timedelta"]=con_time.map(lambda x:x.seconds/60)
        
        dat["start_value"] = data.loc[start_idx]["air_10"].reset_index(drop=True)  
        dat["single_strength"] = env_strength
        
        dat["month"]=dat["starttime"].map(lambda x:x.month)
        dat["weekday"]=dat["starttime"].map(lambda x:x.weekday())
        dat["hour"]=dat["starttime"].map(lambda x:x.hour)
        dat["dev"] = dev  
        allevent=allevent.append(dat)
        	
        dd=dat.dropna()
        dd=dd[dd.timedelta>time_delta]
        dd=dd[dd.single_strength>strength]
        
        # print(dev,len(dat.dropna()))
    
    # sevent=allevent.dropna()
    if out==True:
        if outpath==None:
            allevent.to_csv("gatherlongevent.csv")
        else:
            allevent.to_csv(outpath+"gatherlongevent.csv")
    return allevent

def wavelet_detect(datapack, **kwarg):
    '''Event detection based on wavelet filtered threshold.
    Author: Jiayi Huang & boting Lin
    Parameters
    ----------
        datapack: DataPack
        columns: str or list, default: 'data0'
            which data columns should we processed on. if 'ALL' is given, all data columns will be considered.
        outpath: str, dafault: None
            the path to save the result table. Disable when given None
            
    Return
    ------
        DataFrame, DataFrame
            The detected short event and long event will be returned as a table.
    Note
    ------
        1. Early version, we need more detailed test.
        2. make kwargs for adjustable parameters.
        3. Deal with massive 'value set on copy' warning. and log(0) numerical warning.
        4. function threshold doesnt make sense.
        5. better visualization. at least a report.
        6. how do result depends on the sample frequency.
    '''
    # Parameter Check
    if 'columns' not in kwarg.keys():
        kwarg['columns'] = 'data0'
    if 'outpath' not in kwarg.keys():
        kwarg['outpath'] = None
    if 'date' not in datapack.data.columns:
        datapack.data["date"] = datapack.data.index.map(lambda x:x.date())
    # Main
    shortevent=pd.DataFrame()
    longtermdata=pd.DataFrame()
    longevent=pd.DataFrame()
    for _,site in enumerate(datapack.site_info['site_name']):
        dat = datapack.data[datapack.data['site_name']==site].copy()
        dgroups = groups(dat,'date')
        affiltdata = pd.DataFrame()
        for dgroup in dgroups:
            _, dg, da = group_info(dgroup,'data0') # dg: full table of date, da: dg[air_type]
            a,d=wavelet_decomp(da,5)
            theta=get_theta(dg, 'data0')
            threshold(d,theta)
            ra,rd=wavelet_comp(a,d) # ra: reconstructed a
            dg=after_filter(dg,ra,rd)
            affiltdata=affiltdata.append(dg)

        smooth(dat,affiltdata,site) # moving average of window width 60 and write it into dat
        
        shortevent=shortevent.append(short_event(dat, 'data0'))
        longtermdat=longterm_data(dat,site)
        longtermdata[site]=longtermdat["air_10"]
        longevent=longevent.append(long_event(longtermdat, site))

    gather_long = big_scale(longevent,10,10,out=kwarg['outpath'] is not None,outpath=kwarg['outpath'])   
    gather_short = small_scale(shortevent,out=kwarg['outpath'] is not None,outpath=kwarg['outpath'])
    
    return gather_short, gather_long